/******************************************************************************* 
 * main.cpp - SNES9x PS3
 *
 *  Created on: 2010
 *      Author: Squarepusher2
 *	Author: Themaister
********************************************************************************/
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/timer.h>
#include <sys/return_code.h>
#include <sys/process.h>
#include <cell/sysmodule.h>
#include <cell/cell_fs.h>
#include <cell/pad.h>
#include <cell/dbgfont.h>
#include <cell/audio.h>
#include <stddef.h>
#include <math.h>
#include <sysutil/sysutil_sysparam.h>
#include <sysutil/sysutil_msgdialog.h>
#include <sys/spu_initialize.h>

#include "conffile.h"
#include "main.h"
#include "menu.h"

#ifdef PS3_PROFILING
#include "cellframework/network-stdio/net_stdio.h"
#endif

#include "cellframework/logger/Logger.h"
#include "cellframework/input/cellInput.h"
#include "SNES9xGraphics.h"
#include "emulator_implementation.h"
#include "cellframework/utility/OSKUtil.h"

SYS_PROCESS_PARAM(1001, 0x10000);

// mode the main loop is in
Emulator_Modes mode_switch = MODE_MENU;

// is a ROM running
bool is_running;

// is emulator loaded?
bool emulator_loaded = false;

// needs settings loaded
bool load_settings = true;

// need to load the current rom
bool need_load_rom = false;

// current rom being emulated
char* current_rom = NULL;

CellInputFacade* CellInput;
SNES9xGraphics* Graphics;
Audio::Stream<int16_t> *CellAudio;
OSKUtil *oskutil;

// controller mode to run emulator in
static Emulator_ControllerModes controller_settings = NO_JOYSTICKS;

// debug text for output during emulation loop
const char *debug_text;

float Emulator_GetFontSize()
{
	return Settings.PS3FontSize/100.0;
}

int Emulator_LoadSettings()
{
	return 1;
}

int Emulator_SaveSettings()
{
	if(Emulator_Implementation_SaveSettings())
	{
		return 1;
	}
	else
	{
		return 0;
	}

}


void Emulator_SetControllerMode(Emulator_ControllerModes m)
{
	controller_settings = m;
}


Emulator_ControllerModes Emulator_GetControllerMode()
{
	return controller_settings;
}


bool Emulator_IsInitialized()
{
	return emulator_loaded;
}


bool Emulator_IsROMLoaded()
{
	return current_rom != NULL && need_load_rom == false;
}


bool Emulator_ROMRunning()
{
	return is_running;
}

char * Emulator_GetROM()
{
	return current_rom;
}

void Emulator_SetDebugMessage(const char * value)
{
	debug_text = value;
}

void Emulator_StopAudio()
{
   CellAudio->pause();
}

void Emulator_PlayAudio()
{
   CellAudio->unpause();
}

void Emulator_SwitchMode(Emulator_Modes m)
{
	mode_switch = m;
}


bool Emulator_OSKStart(const wchar_t* msg, const wchar_t* init)
{
	int ret = oskutil->Start(msg,init);
	//FIXME: Does not work?
	/*
	if(ret != 1)
	{
		delete oskutil;
		cellSysmoduleUnloadModule(CELL_SYSMODULE_SYSUTIL_OSK_EXT);
		oskutil = new OSKUtil();
	}
	*/
	return ret;
}

void Emulator_OSKClose()
{
	oskutil->Close();
}

bool Emulator_OSKIsRunning()
{
	return oskutil->IsRunning();
}

const char * Emulator_OSKOutputString()
{
	return oskutil->OutputString();
}

void Emulator_Shutdown()
{
	Emulator_SaveSettings();
#ifdef PS3_PROFILING
   // When profiling we have to do exit(), so that gcov can be hooked properly. sys_process_exit() bypasses all this.
	net_stdio_enable(1);
	LOG_DBG("delete Graphics\n");
	delete Graphics;
	LOG_DBG("delete CellAudio\n");
	delete CellAudio;
	LOG_DBG("delete oskutil\n");
	delete oskutil;

	cellSysmoduleUnloadModule(CELL_SYSMODULE_FS);
	cellSysmoduleUnloadModule(CELL_SYSMODULE_IO);
	cellSysutilUnregisterCallback(0);


	// do any clean up... save stuff etc
	// ...

	// force exit
	exit(0);
#endif

	sys_process_exit(0);
}


void Emulator_RequestLoadROM(char* rom, bool forceReload)
{
	if (forceReload || current_rom == NULL || strcmp(rom, current_rom) != 0)
	{
		if (current_rom != NULL)
		{
			free(current_rom);
		}

		current_rom = strdup(rom);
		need_load_rom = true;
	}
	else
	{
		need_load_rom = false;
	}
}

void Emulator_StopROMRunning()
{
	is_running = false;
}

void Emulator_StartROMRunning()
{
	Emulator_SwitchMode(MODE_EMULATION);
}

/*
 * Callback for PS3 System operations
 */
void sysutil_exit_callback (uint64_t status, uint64_t param, void *userdata)
{
	(void) param;
	(void) userdata;

	switch (status)
	{
		case CELL_SYSUTIL_REQUEST_EXITGAME:
			MenuStop();
			Emulator_StopROMRunning();
			Emulator_SwitchMode(MODE_EXIT);
			break;
		case CELL_SYSUTIL_DRAWING_BEGIN:
		case CELL_SYSUTIL_DRAWING_END:
			break;
		case CELL_SYSUTIL_OSKDIALOG_LOADED:
			LOG_DBG("CELL_SYSUTIL_OSKDIALOG_LOADED\n");
			break;
		case CELL_SYSUTIL_OSKDIALOG_FINISHED:
			LOG_DBG("CELL_SYSUTIL_OSKDIALOG_FINISHED\n");
			oskutil->Stop();
			break;
		case CELL_SYSUTIL_OSKDIALOG_UNLOADED:
			LOG_DBG("CELL_SYSUTIL_OSKDIALOG_UNLOADED\n");
			break;
		default:
			break;
	}
}


bool Emulator_InitSettings ()
{
	Emulator_Implementation_InitSettings();
}


bool Emulator_Init()
{
	return Emulator_Implementation_InitSystem();
}

bool Emulator_Initialize()
{
	emulator_loaded = Emulator_Init();
}


void Emulator_Start()
{
	if(need_load_rom)
	{
		if (Emulator_Initialize())
		{
			need_load_rom = false;
		}
	}

	
	Emulator_Implementation_MuteSound(false);

	//bring down dbgfont
	Graphics->DeInitDbgFont();
	LOG_DBG("Settings.PAL: %d\n", Settings.PAL);
	if (Graphics->GetCurrentResolution() == CELL_VIDEO_OUT_RESOLUTION_576)
	{
		if(Graphics->CheckResolution(CELL_VIDEO_OUT_RESOLUTION_576))
		{
			if(Settings.PAL)
			{
				//Game is PAL
				if(Graphics->GetPAL60Hz())
				{
					//PAL60 is ON, turn it off for PAL
					Emulator_Implementation_SwitchPAL60Hz(false);
				}
			}
			else
			{
				//Game is NTSC
				if(!Graphics->GetPAL60Hz())
				{
					//PAL60 is OFF, turn it on for NTSC
					Emulator_Implementation_SwitchPAL60Hz(true);
				}
			}
		}
	}
	if(Settings.Throttled)
	{
		Emulator_PlayAudio();
	}
	while (is_running)
	{
		Emulator_Implementation_Mainloop();
	}

	Graphics->InitDbgFont();
	Emulator_Implementation_SaveSRAM();
	Emulator_Implementation_MuteSound(true);
	is_running = true;
}

int main (void) {
	sys_spu_initialize(6, 1);
	cellSysmoduleLoadModule(CELL_SYSMODULE_FS);
	cellSysmoduleLoadModule(CELL_SYSMODULE_IO);

	cellSysutilRegisterCallback(0, sysutil_exit_callback, NULL);
   	LOG_INIT();
   	LOG_DBG("LOG INIT\n");

#ifdef PS3_PROFILING
	net_stdio_set_target(PS3_PROFILING_IP, PS3_PROFILING_PORT);
	net_stdio_set_paths("/", 2);
	net_stdio_enable(0);
#endif
	if (load_settings)
	{
		Emulator_InitSettings();
		load_settings =  false;
	}

	LOG_DBG("Graphics = new SNES9xGraphics(%d, %d, %s, %d, %f, %d)\n", Settings.PS3KeepAspect, Settings.PS3Smooth,Settings.PS3CurrentShader.c_str(),Settings.PS3OverscanEnabled, (float)Settings.PS3OverscanAmount/100, Settings.PS3PALTemporalMode60Hz, Settings.Throttled);

	Graphics = new SNES9xGraphics((uint32_t)Settings.PS3CurrentResolution, Settings.PS3KeepAspect, Settings.PS3Smooth,Settings.PS3CurrentShader.c_str(),Settings.PS3OverscanEnabled, (float)Settings.PS3OverscanAmount/100, Settings.PS3PALTemporalMode60Hz, Settings.Throttled);
   	LOG_DBG("Graphics->Init()\n");
	Graphics->Init();
	
	CellInput = new CellInputFacade();
   	LOG_DBG("CellInput->Init()\n");
	CellInput->Init();

	oskutil = new OSKUtil();


	// FIXME: Is this necessary?
	if (Graphics->InitCg() != CELL_OK)
	{
		LOG_DBG("Failed to InitCg: %d\n", __LINE__);
		Emulator_Shutdown();
	}

	LOG_DBG("Graphics->InitDbgFont()\n");
	Graphics->InitDbgFont();

	LOG_DBG("Emulator_Implementation_ToggleSound\n");
	Emulator_Implementation_ToggleSound();
   
	is_running = true;

	while(1)
	{
		switch(mode_switch)
		{
			case MODE_MENU:
				MenuMainLoop();
				break;
			case MODE_EMULATION:
				Emulator_Start();
				if(Settings.Throttled)
				{
					Emulator_StopAudio();
				}
				break;
			case MODE_EXIT:
				Emulator_Shutdown();
				return 0;
		}
	}
}

